/*
    Rational.java
    gvpr ws99/2ooo, bungsblatt 6

    Copyright (C) 1999  Thomas Kindler (tk)
    e-mail: thomas.kindler@gmx.de

    todo: -

    20.11.99: tk, sourcecode aufgerumt.
    19.11.99: tk, initial version.

    Copyright (c)1999,2000 by Thomas Kindler, thomas.kindler@gmx.de

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the
    full License at http://www.gnu.org/copyleft for more details.
*/

class Rational {

  // Zhler und Nenner
  //
  private long  z, n;


  // ggT(a,b) mit Euklidischem Algorithmus bestimmen
  //
  private long ggT( long a, long b ) 
  {
    while (b!=0)  {          
      long r = a % b;              
      a = b;
      b = r;
    }   
    return a;    
  }

  // Auf kleinsten Nenner bringen, und
  // ggf. Nenner positiv machen.
  //
  private void normalize() 
  {
    long ggt = ggT( z, n );
    z /= ggt;
    n /= ggt;
    
    if (n < 0) {
      z = -z;
      n = -n;
    }
  }


  // Standard-Konstruktor, setzt wert auf 0 ein-tel...
  //
  public Rational() 
  {
    z = 0;
    n = 1;
  }

  // Copy-Konstruktor, erzeugt eine Rational-zahl
  // aus einer vorgegebenen.
  //
  public Rational( Rational val ) 
  {
    z = val.z;
    n = val.n;  
  }

  // Konstruktor mit Zhler und Nenner.
  // Prft, ob der Nenner!=0 ist, und 
  // krzt anschlieend den Wert.
  //
  public Rational( long z, long n ) 
  {
    if (n == 0)  
      throw new ArithmeticException( "nenner == 0" );
    
    this.z = z;
    this.n = n;
    
    normalize();
  }

  
  //  a   c     ac
  //  - * -  = ----
  //  b   d     bd
  //
  public Rational multiply( Rational val ) 
  {
    return new Rational( z*val.z, n*val.n );
  }

  //  a   c     ad
  //  - / -  = ----
  //  b   d     bc
  //
  public Rational divide( Rational val ) 
  {
    return new Rational( z*val.n, n*val.z );
  }

  //  a   c     ad+cb
  //  - + -  = -------
  //  b   d      b*d
  //  
  public Rational add( Rational val ) 
  {
    return new Rational( z*val.n + n*val.z, n*val.n );
  }

  //  a   c     ad+cb
  //  - - -  = -------
  //  b   d      b*d
  //  
  public Rational subtract( Rational val ) 
  {
    return new Rational( z*val.n - n*val.z, n*val.n );
  }


  // In String umwandeln, "(z/n)"
  // (berschreibt das von Object geerbte toString() )
  //
  public String toString() 
  {
    return "("+z+"/"+n+")";
  }
 
}