/*  Rechner_Interface.java
    Uni-Do, gvpr 99/2000, Gruppe 22, Blatt 10

    Copyright (C) 2000  Thomas Kindler (tk)
    e-mail: thomas.kindler@gmx.de

    todo: -

    09.01.00: tk, Expression als Interface implementiert.
    02.12.99: tk, initial version.
                                                            
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the 
    full License at http://www.gnu.org/copyleft for more details.    
*/


interface Expression {
  // Berechnet den Wert einer Expresion
  //
  abstract public double calc();
}


class Number implements Expression  {
  
  private double value;
  
  public Number( double v )  { 
    value = v;     
  }
  
  public double calc()  { 
    return value;  
  }
}


abstract class UnOp implements Expression  {
  
  private Expression  exp;

  public UnOp( Expression arg )  {  
    exp = arg;  
  }
  
  // liefert das Argument zurck
  //
  public Expression GetArgument()  {
    return exp;
  }
}


abstract class BinOp implements Expression  {

  private Expression  exp1, exp2;
   
  public BinOp( Expression arg1, Expression arg2 ) {
    exp1 = arg1;
    exp2 = arg2;
  }
  
  // liefert die Argumente zurck
  //
  public Expression GetArgument1()  {
    return exp1;
  }
  
  public Expression GetArgument2()  {
    return exp2;
  }
}


class Sinus extends UnOp  {
  
  public Sinus( Expression arg )  {
    super(arg);  
  }

  // berechnet den Sinus von arg
  //
  public double calc()  {
    return Math.sin( GetArgument().calc() ); 
  }
}


class Cosinus extends UnOp  {
  
  public Cosinus( Expression arg )  {
    super(arg);  
  }

  // berechnet den Cosinus von arg
  //
  public double calc()  {
    return Math.cos( GetArgument().calc() ); 
  }
}


class Plus extends BinOp  {
  
  public Plus( Expression arg1, Expression arg2 )  {
    super( arg1, arg2 );    
  }

  public double calc()  {    
    return GetArgument1().calc() + GetArgument2().calc();
  }
}


class Minus extends BinOp  {
  
  public Minus( Expression arg1, Expression arg2 )  {
    super( arg1, arg2 );    
  }

  public double calc()  {    
    return GetArgument1().calc() - GetArgument2().calc();
  }
}


class Mult extends BinOp  {
  
  public Mult( Expression arg1, Expression arg2 )  {
    super( arg1, arg2 );    
  }

  public double calc()  {    
    return GetArgument1().calc() * GetArgument2().calc();
  }
}


class Div extends BinOp  {
  
  public Div( Expression arg1, Expression arg2 )  {
    super( arg1, arg2 );    
  }

  public double calc()  {    
    return GetArgument1().calc() / GetArgument2().calc();
  }
}



class Rechner  {

  public static void main( String argv[] )  {
    
    System.out.println( "3.5 + 4.5 = " + 
                        new Plus( 
                          new Number( 3.5 ), 
                          new Number( 4.5 )
                        ).calc() 
                      );
 
    System.out.println( "sin(1.59)+1.0 = " + 
                        new Plus( 
                          new Sinus( new Number( 1.59 ) ), 
                          new Number( 1.0 )
                        ).calc() 
                      ); 
  
    System.out.println( "sin((cos(0.0)+1.0)*3.14) = " + 
                        new Sinus( 
                          new Mult( 
                            new Plus( 
                              new Cosinus( new Number( 0.0 ) ), 
                              new Number( 1.0 ) 
                            ), 
                            new Number( 3.14 ) 
                          )
                        ).calc()
                      );
  
    System.out.println( "cos((2.7/sin(4))+3)-(5*sin(cos(sin(7)))) = " +
                        new Minus(                          
                          new Cosinus(
                            new Plus(                              
                              new Div(
                                new Number(2.7),
                                new Sinus( new Number(4) )
                              ),                              
                              new Number(3)
                            )
                          ),
                          new Mult(
                            new Number(5),
                            new Sinus(
                              new Cosinus(
                                new Sinus( new Number(7) )
                              )
                            )                               
                          )                        
                        ).calc()
                      );
  }
}

/* Ausgabe:

    3.5 + 4.5 = 8.0
    sin(1.59)+1.0 = 1.999815615134291
    sin((cos(0.0)+1.0)*3.14) = -0.0031853017931379904
    cos((2.7/sin(4))+3)-(5*sin(cos(sin(7)))) = -2.7150510591347397
*/
