/*  Person.java
    Uni-Do, gvpr 99/2000, Gruppe 22, Blatt 10

    Copyright (C) 2000  Thomas Kindler (tk)
    e-mail: thomas.kindler@gmx.de

    todo: -

    12.01.2000: tk, fr die Lsungsabgabe alles in ein 
                    einziges, groes .java file gepackt.
    10.01.2000: tk, neuer check eingebaut: tote Personen 
                    sollten sich nicht selbst beerben..
    08.01.2000: tk, initial version.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the 
    full License at http://www.gnu.org/copyleft for more details.    
*/

// Familie benutzt eine LinkedList
//
import java.util.*;

// Exception-Hierarchie
//
class  PersonException     extends  RuntimeException  {
  public PersonException()          {  super();   }
  public PersonException(String s)  {  super(s);  }
}

      class  KeinEhepaar  extends  PersonException  {
        public KeinEhepaar()             {  super();   }
        public KeinEhepaar(String s)     {  super(s);  }
      }

      class  PersonIstTot    extends  PersonException  {
        public PersonIstTot()            {  super();   }
        public PersonIstTot(String s)    {  super(s);  }
      }


class  FamilienException     extends  RuntimeException  {
  public FamilienException()             {  super();   }
  public FamilienException(String s)     {  super(s);  }
}

      class  Polygamie       extends  FamilienException  {
        public Polygamie()               {  super();   }
        public Polygamie(String s)       {  super(s);  }
      }

      class  HomoHeirat      extends  FamilienException  {
        public HomoHeirat()              {  super();   }
        public HomoHeirat(String s)      {  super(s);  }
      }

      class  DoppelterName      extends  FamilienException  {
        public DoppelterName()           {  super();   }
        public DoppelterName(String s)   {  super(s);  }
      }

      class  KindNichtGefunden      extends  FamilienException  {
        public KindNichtGefunden()           {  super();   }
        public KindNichtGefunden(String s)   {  super(s);  }
      }



class Person {
  // Einige Konstanten fr den Familienstand
  //  
  public static final int      LEDIG       = 1;
  public static final int      VERHEIRATET = 2;
  public static final int      VERWITWET   = 3;
  public static final boolean  MAENNLICH   = true;
  public static final boolean  WEIBLICH    = false;
 
  private String    name;
  private boolean   geschlecht;
  private boolean   lebendig;
  private double    vermoegen;
  
  private Familie   famGeburt;
  private Familie   famHeirat;
    

  public  Person( String name, boolean geschlecht, Familie eltern )  {
    this.name       = name;
    this.geschlecht = geschlecht;
    this.lebendig   = true;
    this.vermoegen  = 0;

    this.famGeburt  = eltern;
    this.famHeirat  = null;
  }
  
  // Methoden um die Attribute abfragen zu knnen.
  //
  public  boolean  istLebendig()   {  return  lebendig;    }
  public  boolean  gibGeschlecht() {  return  geschlecht;  }
  public  String   gibName()       {  return  name;        }
  public  double   gibVermoegen()  {  return  vermoegen;   }
  public  Familie  gibFamGeburt()  {  return  famGeburt;   }
  public  Familie  gibFamHeirat()  {  return  famHeirat;   }

  public  int      gibFamilienStand()  {
    if (famHeirat == null)
      return  LEDIG;

    if ( famHeirat.gibMann().istLebendig() &&
         famHeirat.gibFrau().istLebendig() )
      return  VERHEIRATET;
    else
      return  VERWITWET;
  }

  // Neue Familie grnden, Regeln werden 
  // im Familienkonstruktor berprft.
  //
  public  void  Heiraten( Person partner )  {
    famHeirat = new Familie( this, partner );
    partner.famHeirat = this.famHeirat;
  }
   

  public  void  KindKriegen( String name, boolean geschlecht )   {    
    if ( gibFamilienStand() != VERHEIRATET )
      throw new KeinEhepaar( name + " ist nicht verheiratet." );

    famHeirat.NeuesKind( new Person( name, geschlecht, famHeirat ) );
  }

  
  public  void  Verdienen( double verdienst )  {
    if (!lebendig)
      throw new PersonIstTot( name + " kann nichts mehr verdienen." );
    
    vermoegen += verdienst;
  }
  

  public  void  Sterben()   {
    if (!lebendig)
      throw new PersonIstTot( name + " kann nur einmal sterben." );

    lebendig = false;
    Vererben();
  }


  public String toString()  {

    String s = "\"" + name + "\", " + 
               (lebendig   == true      ? ""          : "tod, "   ) + 
               (geschlecht == MAENNLICH ? "maennlich" : "weiblich") ;
    
    if (lebendig)  {
      switch (gibFamilienStand())  {
 
        case LEDIG:        s+= ", ledig, ";        break;
        case VERHEIRATET:  s+= ", verheiratet, ";  break;
        case VERWITWET:    s+= ", verwitwet, ";    break;    
      }
      s += vermoegen + " DM";
    }

    return s;
  }


  private void    Vererben()   {
    Person  partner     = null;   // ggf. Ehepartner
    int     kinder      = 0;      // anzahl erbender Kinder
    int     geschwister = 0;      // anzahl erbender Geschwister

    if (famHeirat != null)  {
      // Wenn es eine Familie gibt, so wird der Ehepartner
      // (sofern dieser noch lebt), und danach die Anzahl der 
      // erbfhigen (dh. lebenden) Kinder ermittelt.
      // 
      partner = (geschlecht == MAENNLICH) ? 
                famHeirat.gibFrau() : famHeirat.gibMann();

      if (!partner.lebendig)
        partner = null;

      // Kinder durchzhlen
      //
      Iterator  i = famHeirat.gibKinder();
      while (i.hasNext())
        if ( ((Person)i.next()).istLebendig() )
          kinder++;
    }


    if (famGeburt != null)  {
      // Falls wir nicht gerade Adam oder Eva sind,
      // wird geprft wieviele lebende geschwister 
      // es gibt.
      //  
      Iterator  i = famGeburt.gibKinder();
      while (i.hasNext()) {
        Person p = (Person)i.next();
        if ( p.lebendig  &&  p != this )
          geschwister++;
      }
    }


    if (partner != null)  {
      if (kinder > 0)  {
        // Der Partner erhlt 50%, der Rest 
        // wird auf die Kinder verteilt.
        // 
        partner.Verdienen( vermoegen / 2 );

        Iterator i = famHeirat.gibKinder();
        while (i.hasNext())  {
          Person p = (Person)i.next();
          if (p.lebendig)
            p.Verdienen( (vermoegen/2) / kinder );
        }
      } else  {
        // Ohne Kinder erhlt der Partner 100%.
        //
        partner.Verdienen( vermoegen );
      }    
    }  else  {

      if (kinder > 0)  {
        // Wenn es keinen Partner gibt, wird unter
        // den Kinder aufgeteilt.
        //
        Iterator i = famHeirat.gibKinder();
        while (i.hasNext())  {
          Person p = (Person)i.next();
          if (p.lebendig)
            p.Verdienen( vermoegen / kinder );
        }
      } else {
        // Ansonsten geht das Vermoegen an die Geschwister.
        //
        if (geschwister > 0)  {
          Iterator i = famGeburt.gibKinder();
          while (i.hasNext())  {
            Person p = (Person)i.next();
            if ( p.lebendig  &&  p!=this )
              p.Verdienen( vermoegen / geschwister );
          }
        }
      }  
    
    }

    // In jedem Fall verliert der Tote sein Vermgen.
    //
    vermoegen = 0;  
  }

}



class  Familie  {

  private  Person      mann;
  private  Person      frau;
  private  LinkedList  kinder;

  public   Iterator    gibKinder()  { return kinder.iterator(); };
  public   Person      gibMann()    { return mann;  };
  public   Person      gibFrau()    { return frau;  };


  // Erzeugt ein neues Familienobjekt, dabei 
  // werden auch die Heiratsregeln berprft.  
  //
  public  Familie( Person p1, Person p2 )   {    
    if ( !p1.istLebendig() || !p2.istLebendig() )
      throw new PersonIstTot( "Heirat mit toten nicht erlaubt." );

    if ( !((p1.gibGeschlecht() == Person.MAENNLICH)^ 
           (p2.gibGeschlecht() == Person.MAENNLICH)) )
      throw new HomoHeirat( "Homo-Ehen sind nicht erlaubt." );
                
    if ( (p1.gibFamilienStand() != Person.LEDIG) ||
         (p2.gibFamilienStand() != Person.LEDIG) )
      throw new Polygamie( "Polygamie ist nicht erlaubt." );    

    if ( p1.gibGeschlecht() == Person.MAENNLICH ) {
      mann  = p1;
      frau  = p2;
    } else {
      mann  = p2;
      frau  = p1;
    }
    
    kinder = new LinkedList();
  }


  public  void  NeuesKind( Person kind )   {
    try {
      // falls SucheKind() ein Kind mit diesem Namen
      // findet, wird eine Exception geworfen.
      //
      SucheKind( kind.gibName() );
      throw new DoppelterName( kind.gibName() + " gibt es schon." );
    }  catch (KindNichtGefunden e) {
      // Ansonsten wird das Kind hinzugefgt.
      //      
      kinder.add( kind );
    }  
  }

 
  // Sucht nach einem Kind anhand seines Namens.
  //
  public  Person  SucheKind( String name )   {
    Iterator i = gibKinder();
    while (i.hasNext())  {
      
      Person  p = (Person)i.next();
      if ( p.gibName().equals( name ) )
        return p;
    }  
    throw new KindNichtGefunden( "kann " + name +" nicht finden." );
  }


  public  String  toString()  {
    
    String s = "Familie \"" + mann.gibName() + 
               " & " + frau.gibName() + "\" ----\n" + 
               "  Mann: " + mann + "\n" + 
               "  Frau: " + frau + "\n";

    s += "  " + kinder.size() + " Kind(er)\n"; 

    Iterator  i = gibKinder();
    while ( i.hasNext() )
      s += "  - " + (Person)i.next() + "\n";

    return s;
  }

}



class  Test  {
  
  public static void main(String[] argv)  {
  
    // Erzeugt Adam und Eva, lt sie ein bisschen Geld 
    // verdienen, und schlielich heiraten...
    //
    Person  adam = new Person( "Adam", Person.MAENNLICH, null );
    Person  eva  = new Person( "Eva" , Person.WEIBLICH,  null );

    adam.Verdienen( 3000 );
    eva.Verdienen( 1500 );
    adam.Heiraten( eva );

    // Jetzt drfen die beiden auch ganz offiziell rumpoppen :)  
    // Die Exceptions werden mit e.printStackTrace() ausgegeben,
    // damit man sieht, wo sie geworfen wurden.
    //
    try {
      eva.KindKriegen( "Arthur",   Person.MAENNLICH );
      eva.KindKriegen( "Ford",     Person.MAENNLICH );
      eva.KindKriegen( "Zaphod",   Person.MAENNLICH );
      eva.KindKriegen( "Trillian", Person.WEIBLICH  );
      eva.KindKriegen( "Trillian", Person.WEIBLICH  );
    } catch  (Exception e)  {  e.printStackTrace();  }

    Person  arthur, ford, zaphod, trillian, marvin;

    Familie  adamUndEva = adam.gibFamHeirat();
    arthur   =  adamUndEva.SucheKind( "Arthur" );
    ford     =  adamUndEva.SucheKind( "Ford" );
    zaphod   =  adamUndEva.SucheKind( "Zaphod" );
    trillian =  adamUndEva.SucheKind( "Trillian"  );
    try  {
      marvin   =  adamUndEva.SucheKind( "Marvin" );
    } catch (Exception e)  {  e.printStackTrace();  }
    
    try  {
      trillian.Heiraten( zaphod );
      trillian.KindKriegen( "Eddie", Person.MAENNLICH );
      
      zaphod.Heiraten( arthur );
    } catch (Exception e)  {  e.printStackTrace();  } 
    
    try {
      zaphod.Verdienen( 10000 );
      trillian.Verdienen( 2000 );
      
      arthur.Heiraten( trillian );

    } catch (Exception e)  {  e.printStackTrace();  } 

    try {
      adam.Sterben();           // Grippewelle !!
      ford.Verdienen( 6000 );   // Aber Ford macht das nichts aus...
      ford.Sterben();           // oder doch ??
      adam.Sterben();           // das htte auch adam nich gedacht..
    } catch (Exception e)  {  e.printStackTrace();  }
    
    System.out.println( adam.gibFamHeirat() );
    System.out.println( zaphod.gibFamHeirat() );
  }
}

/*  Ausgabe:
    DoppelterName: Trillian gibt es schon.
            at Familie.NeuesKind(Person.java:310)
            at Person.KindKriegen(Person.java:132)
            at Test.main(Person.java:376)
    KindNichtGefunden: kann Marvin nicht finden.
            at Familie.SucheKind(Person.java:329)
            at Test.main(Person.java:387)
    HomoHeirat: Homo-Ehen sind nicht erlaubt.
            at Familie.<init>(Person.java:286)
            at Person.Heiraten(Person.java:123)
            at Test.main(Person.java:394)
    Polygamie: Polygamie ist nicht erlaubt.
            at Familie.<init>(Person.java:290)
            at Person.Heiraten(Person.java:123)
            at Test.main(Person.java:401)
    PersonIstTot: Adam kann nur einmal sterben.
            at Person.Sterben(Person.java:146)
            at Test.main(Person.java:409)
    Familie "Adam & Eva" ----
      Mann: "Adam", tod, maennlich
      Frau: "Eva", weiblich, verwitwet, 3000.0 DM
      4 Kind(er)
      - "Arthur", maennlich, ledig, 2500.0 DM
      - "Ford", tod, maennlich
      - "Zaphod", maennlich, verheiratet, 12500.0 DM
      - "Trillian", weiblich, verheiratet, 4500.0 DM

    Familie "Zaphod & Trillian" ----
      Mann: "Zaphod", maennlich, verheiratet, 12500.0 DM
      Frau: "Trillian", weiblich, verheiratet, 4500.0 DM
      1 Kind(er)
      - "Eddie", maennlich, ledig, 0.0 DM
*/