/*  Mission.java
    gvpr ws99/2ooo, bungsblatt 9

    Copyright (C) 2000  Thomas Kindler (tk)
    e-mail: thomas.kindler@gmx.de

    todo: -

    13.01.00: tk, - Die History() Klasse benutzt jetzt ein BitSet, um
                    die Zustnde zu speichern.  Der Index wird mit
                    Zustand.toIntegerIndex() gebildet. Mein PII450
                    schafft (1000,999,L,0,0) in ca. 43 sekunden...
                  - History() wirft NoWayException.
                  - Es wird ein eindeutiger Lsungsweg ausgegeben.
    10.01.00: tk, source komplett berarbeitet.
    08.01.00: tk, wenigstens luft es berhaupt :)..
    16.12.99: tk, initial version.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the
    full License at http://www.gnu.org/copyleft for more details.
*/

// History benutzt ein BitSet
//
import java.util.*;

class  ZustandsException  extends  Exception  {
  ZustandsException()            { super();  }
  ZustandsException( String s )  { super(s); }
};

class  NoWayException     extends  Exception  {
  NoWayException()               { super();  }
  NoWayException( String s )     { super(s); }
};


class Zustand  {

  public static final boolean LINKS  = false;
  public static final boolean RECHTS = true;

  private  int      ml, kl;   // missionare, kannibalen links
  private  int      mr, kr;   //   "    "      "    "   rechts
  private  boolean  boot;     // bootposition

  // Der Zustand wird im Konstruktor auf Gltigkeit
  // und Ungefhrlichkeit berprft, ggF. wird eine
  // ZustandsException geworfen.
  //
  public  Zustand( int ml, int kl, boolean boot, int mr, int kr )
          throws  ZustandsException
  {
    this.ml   = ml;
    this.kl   = kl;
    this.boot = boot;
    this.mr   = mr;
    this.kr   = kr;

    if (!istGueltig())
      throw  new  ZustandsException( this + " ist ungltig." );

    if (!istUngefaehrlich())
      throw  new  ZustandsException( this + " ist gefhrlich." );
  }

  // Ueberfahrt() berprueft die Bootbesetzung,
  // und liefert dann den neuen Zustand zurck.
  //
  public  Zustand  Ueberfahrt( int m, int k )
          throws   ZustandsException  
  {
    if ( (m <  0  ||  k <  0) ||
         (m == 0  &&  k == 0) ||
         (m + k > 2) )
      throw new ZustandsException(
                  this + ": Ueberfahrt("+m+","+k+") nicht mglich."
                );

    if (boot == LINKS)
      return new Zustand( ml-m, kl-k, RECHTS, mr+m, kr+k );
    else
      return new Zustand( ml+m, kl+k, LINKS,  mr-m, kr-k );
  }

  // Das Ziel ist erreicht, wenn keiner links
  // steht, und das Boot am rechten Ufer liegt.
  //
  public   boolean  istZielZustand() {
    return ml == 0  &&  kl == 0  &&  boot == RECHTS;
  }

  // toIntegerIndex() liefert eine injektive Abbildung 
  // des Zustandes auf einen Integer-Wert, maxIndex() 
  // liefert den dabei grtmglichen Index. Es gilt:
  // 0  <=  toIntegerIndex()  <  maxIndex() 
  //
  public  int  toIntegerIndex()  {
    return  (boot == LINKS ? 0 : 1) + (ml * 2) + (kl * 2*(ml+mr+1));
  }

  public  int  maxIndex()  {
    return  2*(ml+mr+1)*(kl+mr+1);
  }

  private  boolean  istGueltig()  {
    return  ml >= 0  &&  kl >= 0  &&
            mr >= 0  &&  kr >= 0;
  }

  private  boolean  istUngefaehrlich()  {
    return  (ml >= kl  ||  ml == 0) &&
            (mr >= kr  ||  mr == 0);
  }

  public  String  toString()  {
    return  "(" + ml + ", " + kl + ", " +
            (boot == LINKS ? "L" : "R") +
            ", " + mr + ", " + kr + ")" +
            (istZielZustand() ? " Ziel" : "");
  }
}


class FolgeZustaende  {

  private int     anzahl  = 0;
  private Zustand liste[] = new Zustand[5];
  private static  final int fahrten[][] = {{0,2}, {0,1}, {1,1}, {1,0}, {2,0}};

  public FolgeZustaende( Zustand zustand )  {

    for (int i=0; i<liste.length; i++)
      try {
        liste[anzahl] = zustand.Ueberfahrt( fahrten[i][0],
                                            fahrten[i][1] );
        anzahl++;
      }  catch  (ZustandsException e)  {
        // Wenn eine Ueberfahrt eine Exception wirft,
        // wird einfach die nchstmgliche probiert.
        //
      };
  }

  public Zustand  Naechster()
         throws NoWayException
  {
    if  (anzahl == 0)
      throw new NoWayException( "Keine Folgezustnde mehr mglich." );

    return  liste[--anzahl];
  }
}


class History  {

  private  BitSet  liste;

  public   History( int size )  {
    liste = new BitSet( size );
  }

  public  void    add( Zustand zustand )
          throws  NoWayException
  {
    if ( liste.get(zustand.toIntegerIndex()) )
      throw new NoWayException( zustand + " zyklus" );

    liste.set( zustand.toIntegerIndex() );
  }
}


class Control  {

  private  History       history;
  private  StringBuffer  loesungsweg;


  public Control( Zustand zustand )
         throws NoWayException
  {    
    history     = new History( zustand.maxIndex() );
    loesungsweg = new StringBuffer();
  
    TiefenSuche( zustand );
    System.out.println( "Lsung gefunden:\n" + loesungsweg );
  }


  private void    TiefenSuche( Zustand  zustand )
          throws  NoWayException
  {
    history.add( zustand );

    // Wenn der Zielzustand noch nicht erreicht 
    // ist, wird in den Folgezustnden gesucht.
    //
    if (!zustand.istZielZustand())
      TiefenSucheFZ( new FolgeZustaende( zustand ) );

    // Wenn die Untersuchung der Folgezustnde nicht in 
    // eine Sackgasse gefhrt hat, wird der Lsungsweg 
    // eingetragen; dabei wird der callstack schrittweise 
    // bis zur Ausgangssituation abgebaut.
    //
    loesungsweg.insert( 0, "  "+ zustand + "\n" );
  }


  private void    TiefenSucheFZ( FolgeZustaende folgezustand )
          throws  NoWayException
  {
    Zustand zustand = folgezustand.Naechster();

    try {
      TiefenSuche( zustand );
    } catch ( NoWayException e )  {
      TiefenSucheFZ( folgezustand );
    }
  }
}

class Mission  {

  static public void main( String argv[] )  {
    try  {
      Zustand zustand = new Zustand( 3, 3, Zustand.LINKS, 0, 0 );    
      Control control = new Control( zustand );
    }  catch (Exception e)  {
      System.out.println( e );
    }
  }

}

/*  Ausgabe:
      Lsung gefunden:
        (3, 3, L, 0, 0)
        (2, 2, R, 1, 1)
        (3, 2, L, 0, 1)
        (3, 0, R, 0, 3)
        (3, 1, L, 0, 2)
        (1, 1, R, 2, 2)
        (2, 2, L, 1, 1)
        (0, 2, R, 3, 1)
        (0, 3, L, 3, 0)
        (0, 1, R, 3, 2)
        (1, 1, L, 2, 2)
        (0, 0, R, 3, 3) Ziel
*/
