/*  Springer.java
    Lst das Springerproblem von Daniel Chernukhin

    Copyright (C) 2000  Thomas Kindler (tk)
    e-mail: thomas.kindler@gmx.de

    todo: -

    18.01.00: tk, initial version, based on mission.java.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the
    full License at http://www.gnu.org/copyleft for more details.
*/

import java.util.*;

class  NoWayException     extends  Exception  {
  NoWayException()               { super();  }
  NoWayException( String s )     { super(s); }
};


class Zustand  {

  // das Brett ist aus effizienzgrnden (schnelles kopieren!)
  // nur eindimensional. 
  private  boolean  brett[];  
 
  private  int      breite, hoehe;
  private  int      springerX, springerY;
  int               zugNr;

  // Erzeugt ein leeres Schachbrett ohne Springer
  //
  public  Zustand( int b, int h )  {
    breite    = b;
    hoehe     = h;
    springerX = 0;
    springerY = 0;
    zugNr     = 0;
    brett     = new boolean[breite * hoehe];
  }
  
  // Kopiert ein Schachbrett, und fhrt einen Zug aus.
  //
  public  Zustand( int dx, int dy, Zustand other )  
          throws NoWayException
  {
    breite    = other.breite;
    hoehe     = other.hoehe;
    springerX = other.springerX + dx;
    springerY = other.springerY + dy;
    zugNr     = other.zugNr + 1;
    brett     = (boolean[])other.brett.clone();

    if (springerX < 0 || springerX >= breite ||
        springerY < 0 || springerY >= hoehe )
      throw new NoWayException( "Ungltige Position!" );

    if (brett[springerX + springerY*breite])
      throw new NoWayException( "Position schon besetzt!" );

    brett[springerX + springerY*breite] = true;
  }
  
  
  // Das Ziel ist erreicht, wenn alle Felder
  // angesprungen wurden.
  //
  public   boolean  istZielZustand() {
    return zugNr == breite*hoehe;    
  }

  public  String  toString()  {
    String s = "";
    
    for (int y=0; y<hoehe; y++)  {
      for (int x=0; x<breite; x++) {
        if (x==springerX && y==springerY)
          s += "#";
        else
          s += (brett[x + y*breite]) ? "o" : "."; 
      }
      s += "\n";
    }
    s += "\n";
    return s; 
  }
}


class FolgeZustaende  {

  private int       anzahl  = 0;
  private Zustand   liste[] = new Zustand[8];
  private final int zuege[][] = { {+1,+2}, {+1,-2}, {-1,+2}, {-1, -2},
                                  {+2,+1}, {+2,-1}, {-2,+1}, {-2, -1} };

  public FolgeZustaende( Zustand zustand )  {

    for (int i=0; i<liste.length; i++)
      try {
        liste[anzahl] = new Zustand( zuege[i][0], zuege[i][1], zustand );
        anzahl++;
      }  catch  (NoWayException e)  {
        // Wenn ein Zug eine Exception wirft,
        // wird einfach der nchste probiert.
        //
      };
  }

  public Zustand  Naechster()
         throws NoWayException
  {
    if  (anzahl == 0)
      throw new NoWayException( "Keine Folgezustnde mehr mglich." );

    return  liste[--anzahl];
  }
}


class Control  {
  private  int  versuchNr;
  private  int  maxTiefe;

  public Control( Zustand zustand )
         throws NoWayException
  {    
    versuchNr = 0;

    TiefenSuche( zustand );

    System.out.println( "Lsung gefunden!\n" );
  }


  private void    TiefenSuche( Zustand  zustand )
          throws  NoWayException
  {
    // Wenn der Zielzustand noch nicht erreicht 
    // ist, wird in den Folgezustnden gesucht.
    //
    versuchNr++;
    if (maxTiefe < zustand.zugNr) 
      maxTiefe=zustand.zugNr;

    System.out.print( "\rversuch #" + versuchNr + " bisher geloest: " + maxTiefe + 
                      " tiefe: " + zustand.zugNr );
    
    if (!zustand.istZielZustand())
      TiefenSucheFZ( new FolgeZustaende( zustand ) );

    // Wenn die Untersuchung der Folgezustnde nicht in 
    // eine Sackgasse gefhrt hat, wird der Lsungsweg 
    // eingetragen; dabei wird der stack schrittweise 
    // bis zur Ausgangssituation abgebaut.
    //
    System.out.println( zustand );
  }


  private void    TiefenSucheFZ( FolgeZustaende folgezustand )
          throws  NoWayException
  {
    Zustand zustand = folgezustand.Naechster();

    try {
      TiefenSuche( zustand );
    } catch ( NoWayException e )  {
      TiefenSucheFZ( folgezustand );
    }
  }
}


class Springer  {

  static public void main( String argv[] )  {

    try {
      Zustand  z = new Zustand( 5, 5 );
      z = new Zustand( 0, 0, z );

      Control  c = new Control( z );


    } catch (Exception e)  {
      System.out.println( "\nkeine Loesung moeglich!" );
      e.printStackTrace();
    }

  }

}